import asyncio
from functools import wraps

import typer

from .settings import settings

app = typer.Typer()


def syncify(f):
    """This simple decorator converts an async function into a sync function,
    allowing it to work with Typer.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@app.command(help=f"Display the current installed version of {settings.project_name}.")
def version():
    from . import _version

    typer.echo(f"{settings.project_name} - {_version.version}")


if __name__ == "__main__":
    app()



from poromics import julia_helpers

__all__ = ["tortuosity_fd", "_jl", "_taujl"]

julia_helpers.ensure_julia_deps_ready(quiet=False)

_jl = julia_helpers.init_julia(quiet=False)
_taujl = julia_helpers.import_backend(_jl)


class Result:
    """Container for storing the results sof a tortuosity simulation."""

    def __init__(self, im, axis, tau, c):
        """
        Initializes a tortuosity result object.

        Args:
            tau (float): The tortuosity value.
            c_grid (ndarray): The concentration grid solution.
            axis (int or str): The axis along which tortuosity was calculated.

        """
        self.im = im
        self.axis = axis
        self.tau = tau
        self.c = c

    def __repr__(self):
        return f"Result(τ = {self.tau:.2f}, axis = {self.axis})"


def tortuosity_fd(im, *, axis: int, rtol: float = 1e-5, gpu: bool = False) -> Result:
    """
    Performs a tortuosity simulation on the given image along the specified axis."

    Args:
        im (ndarray): The input image.
        axis (int): The axis along which to compute tortuosity (0=x, 1=y, 2=z).
        rtol (float): Relative tolerance for the solver.
        gpu (bool): If True, use GPU for computation.

    Returns:
        Result: An object containing the original image, axis, tortuosity,
        and concentration grid.

    """
    axis_jl = _jl.Symbol(["x", "y", "z"][axis])
    sim = _taujl.TortuositySimulation(im, axis=axis_jl, gpu=gpu)
    sol = _taujl.solve(sim.prob, _taujl.KrylovJL_CG(), verbose=False, reltol=rtol)
    c_grid = _taujl.vec_to_grid(sol.u, im)
    tau = _taujl.tortuosity(c_grid, axis_jl)
    return Result(im, axis, tau, c_grid)



import shutil
from pathlib import Path

import juliapkg
from juliapkg.deps import can_skip_resolve
from juliapkg.find_julia import find_julia
from loguru import _defaults, logger
from tqdm.auto import tqdm

from .utils import suppress_output

# Remove milliseconds from loguru format for cleaner output
logger.remove()
logger_fmt = _defaults.LOGURU_FORMAT
logger_fmt = logger_fmt.replace(".SSS", "")
logger.add(lambda msg: tqdm.write(msg, end=""), format=logger_fmt, colorize=True)


def install_julia(quiet: bool = False):
    """Installs Julia using juliapkg."""
    # Importing juliacall automatically installs Julia using juliapkg
    if quiet:
        with suppress_output():
            import juliacall  # noqa: F401
    else:
        import juliacall  # noqa: F401


def install_backend(quiet: bool = False):
    """Installs Julia dependencies for Poromics.

    Args:
        ec_path: Path to the local copy of Tortuosity.jl. Default is None.
            If None, the remote version will be used.
        quiet: If True, suppresses output during installation. Default is False.

    Raises:
        ImportError: If Julia is not installed.
    """
    is_julia_installed(error=True)

    if quiet:
        with suppress_output():
            juliapkg.resolve()
    else:
        juliapkg.resolve()


def init_julia(quiet: bool = False):
    """Initializes Julia and returns the Main module.

    Args:
        quiet: If True, suppresses the output of Julia initialization. Default is False.

    Returns:
        The Julia Main module.

    Raises:
        ImportError: If Julia is not installed.
    """
    is_julia_installed(error=True)
    if not can_skip_resolve():
        logger.warning("Julia is installed, but needs to be resolved...")
    if quiet:
        with suppress_output():
            from juliacall import Main  # type: ignore
    else:
        from juliacall import Main  # type: ignore

    return Main


def import_package(package_name: str, Main, error: bool = False):
    """Imports a package in Julia and returns the module.

    Args:
        package_name: Name of the Julia package to import.
        Main: Julia Main module.
        error: If True, raises an error if the package is not found. Default is False.

    Returns:
        The imported Julia module.

    Raises:
        ImportError: If the package is not found and error is True.
    """
    from juliacall import JuliaError

    try:
        Main.seval(f"using {package_name}")
        return eval(f"Main.{package_name}")
    except JuliaError as e:
        if error:
            raise e
    return None


def import_backend(Main=None):
    """Imports Tortuosity.jl package from Julia.

    Args:
        Main: Julia Main module. Default is None. If None, the Main module will
        be initialized.

    Returns:
        The imported Julia module.

    Raises:
        ImportError: If Julia is not installed or the package is not found.
    """
    Main = init_julia() if Main is None else Main
    is_backend_installed(Main=Main, error=True)
    return import_package("Tortuosity", Main)


def is_julia_installed(error: bool = False):
    """Checks that Julia is installed."""
    # Look for system-wide Julia executable
    try:
        find_julia()
        return True
    except Exception:
        pass
    # Look for local Julia executable (e.g., installed by juliapkg)
    if can_skip_resolve():
        return True
    msg = "Julia not found. Visit https://github.com/JuliaLang/juliaup and install Julia."
    if error:
        raise ImportError(msg)
    return False


def is_backend_installed(Main=None, error: bool = False):
    """Checks if Tortuosity.jl is installed.

    Args:
        Main: Julia Main module. Default is None. If None, the Main module will be
        error: If True, raises an error if the package is not found. Default is False.

    Returns:
        bool: True if the package is installed, False otherwise.

    Raises:
        ImportError: If Julia is not installed or the package is not found and error is True.
    """
    Main = init_julia() if Main is None else Main
    if import_package("Tortuosity", Main, error=False) is not None:
        return True
    msg = "Tortuosity.jl not found, run 'python -m poromics install'"
    if error:
        raise ImportError(msg)
    return False


def ensure_julia_deps_ready(quiet: bool = False, retry: bool = True):
    """Ensures Julia and Tortuosity.jl are installed."""

    def _ensure_julia_deps_ready(quiet):
        if not is_julia_installed(error=False):
            logger.warning("Julia not found, installing Julia...")
            install_julia(quiet=quiet)
        Main = init_julia(quiet=quiet)
        if not is_backend_installed(Main=Main, error=False):
            logger.warning("Julia dependencies not found, installing Tortuosity.jl...")
            install_backend(quiet=quiet)

    def _reset_julia_env(quiet):
        remove_julia_env()
        if quiet:
            with suppress_output():
                juliapkg.resolve(force=True)
        else:
            juliapkg.resolve(force=True)

    try:
        _ensure_julia_deps_ready(quiet)
    except Exception:
        if retry:
            _reset_julia_env(quiet)
            _ensure_julia_deps_ready(quiet)
            return
        raise


def remove_julia_env():
    """Removes the active Julia environment directory.

    When Julia or its dependencies are corrupted, this is a possible fix.
    """
    path_julia_env = Path(juliapkg.project())

    if path_julia_env.exists():
        logger.warning(f"Removing Julia environment directory: {path_julia_env}")
        shutil.rmtree(path_julia_env)
    else:
        logger.warning("Julia environment directory not found.")



from contextlib import contextmanager
import sys
import os
import io


@contextmanager
def suppress_output():
    """Suppresses stdout and stderr for both Unix and Windows systems."""
    # Save the original high-level streams
    saved_stderr = sys.stderr
    saved_stdout = sys.stdout

    try:
        # Try to save the current file descriptors
        original_stderr_fd = sys.stderr.fileno()
        original_stdout_fd = sys.stdout.fileno()
        saved_stderr_fd = os.dup(original_stderr_fd)
        saved_stdout_fd = os.dup(original_stdout_fd)

        with open(os.devnull, "wb") as devnull:
            # Redirect the lower-level file descriptors
            os.dup2(devnull.fileno(), original_stderr_fd)
            os.dup2(devnull.fileno(), original_stdout_fd)

        # Redirect the higher-level Python streams
        sys.stderr = open(os.devnull, "w")
        sys.stdout = open(os.devnull, "w")

    except (io.UnsupportedOperation, AttributeError):
        # If fileno is not supported, just replace the Python streams
        sys.stderr = open(os.devnull, "w")
        sys.stdout = open(os.devnull, "w")

    try:
        yield
    finally:
        # Restore the high-level Python streams
        sys.stderr.close()
        sys.stdout.close()
        sys.stderr = saved_stderr
        sys.stdout = saved_stdout

        # Restore the original file descriptors if they were saved
        if "saved_stderr_fd" in locals() and "saved_stdout_fd" in locals():
            os.dup2(saved_stderr_fd, original_stderr_fd)
            os.dup2(saved_stdout_fd, original_stdout_fd)
            os.close(saved_stderr_fd)
            os.close(saved_stdout_fd)
